@@ -145,66 +145,6 @@ net_init_with_trie(struct fib *f, void *N)
     trie_add_prefix(tab->trie_new, n->n.addr, n->n.addr->pxlen, n->n.addr->pxlen);
 }
 
-static inline net *
-net_route_ip4_trie(rtable *t, const net_addr_ip4 *n0)
-{
-  TRIE_WALK_TO_ROOT_IP4(t->trie, n0, n)
-  {
-    net *r;
-    if (r = net_find_valid(t, (net_addr *) &n))
-      return r;
-  }
-  TRIE_WALK_TO_ROOT_END;
-
-  return NULL;
-}
-
-static inline net *
-net_route_vpn4_trie(rtable *t, const net_addr_vpn4 *n0)
-{
-  TRIE_WALK_TO_ROOT_IP4(t->trie, (const net_addr_ip4 *) n0, px)
-  {
-    net_addr_vpn4 n = NET_ADDR_VPN4(px.prefix, px.pxlen, n0->rd);
-
-    net *r;
-    if (r = net_find_valid(t, (net_addr *) &n))
-      return r;
-  }
-  TRIE_WALK_TO_ROOT_END;
-
-  return NULL;
-}
-
-static inline net *
-net_route_ip6_trie(rtable *t, const net_addr_ip6 *n0)
-{
-  TRIE_WALK_TO_ROOT_IP6(t->trie, n0, n)
-  {
-    net *r;
-    if (r = net_find_valid(t, (net_addr *) &n))
-      return r;
-  }
-  TRIE_WALK_TO_ROOT_END;
-
-  return NULL;
-}
-
-static inline net *
-net_route_vpn6_trie(rtable *t, const net_addr_vpn6 *n0)
-{
-  TRIE_WALK_TO_ROOT_IP6(t->trie, (const net_addr_ip6 *) n0, px)
-  {
-    net_addr_vpn6 n = NET_ADDR_VPN6(px.prefix, px.pxlen, n0->rd);
-
-    net *r;
-    if (r = net_find_valid(t, (net_addr *) &n))
-      return r;
-  }
-  TRIE_WALK_TO_ROOT_END;
-
-  return NULL;
-}
-
 static inline void *
 net_route_ip6_sadr_trie(rtable *t, const net_addr_ip6_sadr *n0)
 {
@@ -238,69 +178,6 @@ net_route_ip6_sadr_trie(rtable *t, const net_addr_ip6_sadr *n0)
   return NULL;
 }
 
-static inline net *
-net_route_ip4_fib(rtable *t, const net_addr_ip4 *n0)
-{
-  net_addr_ip4 n;
-  net_copy_ip4(&n, n0);
-
-  net *r;
-  while (r = net_find_valid(t, (net_addr *) &n), (!r) && (n.pxlen > 0))
-  {
-    n.pxlen--;
-    ip4_clrbit(&n.prefix, n.pxlen);
-  }
-
-  return r;
-}
-
-static inline net *
-net_route_vpn4_fib(rtable *t, const net_addr_vpn4 *n0)
-{
-  net_addr_vpn4 n;
-  net_copy_vpn4(&n, n0);
-
-  net *r;
-  while (r = net_find_valid(t, (net_addr *) &n), (!r) && (n.pxlen > 0))
-  {
-    n.pxlen--;
-    ip4_clrbit(&n.prefix, n.pxlen);
-  }
-
-  return r;
-}
-
-static inline net *
-net_route_ip6_fib(rtable *t, const net_addr_ip6 *n0)
-{
-  net_addr_ip6 n;
-  net_copy_ip6(&n, n0);
-
-  net *r;
-  while (r = net_find_valid(t, (net_addr *) &n), (!r) && (n.pxlen > 0))
-  {
-    n.pxlen--;
-    ip6_clrbit(&n.prefix, n.pxlen);
-  }
-
-  return r;
-}
-
-static inline net *
-net_route_vpn6_fib(rtable *t, const net_addr_vpn6 *n0)
-{
-  net_addr_vpn6 n;
-  net_copy_vpn6(&n, n0);
-
-  net *r;
-  while (r = net_find_valid(t, (net_addr *) &n), (!r) && (n.pxlen > 0))
-  {
-    n.pxlen--;
-    ip6_clrbit(&n.prefix, n.pxlen);
-  }
-
-  return r;
-}
 
 static inline void *
 net_route_ip6_sadr_fib(rtable *t, const net_addr_ip6_sadr *n0)
@@ -346,164 +223,58 @@ net *
 net_route(rtable *tab, const net_addr *n)
 {
   ASSERT(tab->addr_type == n->type);
+  net_addr_union *nu = SKIP_BACK(net_addr_union, n, n);
 
-  switch (n->type)
-  {
-  case NET_IP4:
-    if (tab->trie)
-      return net_route_ip4_trie(tab, (net_addr_ip4 *) n);
-    else
-      return net_route_ip4_fib (tab, (net_addr_ip4 *) n);
+#define TW(ipv, what) \
+  TRIE_WALK_TO_ROOT_IP##ipv(tab->trie, &(nu->ip##ipv), var) \
+  { what(ipv, var); } \
+  TRIE_WALK_TO_ROOT_END; return NULL;
 
-  case NET_VPN4:
-    if (tab->trie)
-      return net_route_vpn4_trie(tab, (net_addr_vpn4 *) n);
-    else
-      return net_route_vpn4_fib (tab, (net_addr_vpn4 *) n);
+#define FW(ipv, what) do { \
+  net_addr_union nuc; net_copy(&nuc.n, n); \
+  while (1) { \
+    what(ipv, nuc.ip##ipv); if (!nuc.n.pxlen) return NULL; \
+    nuc.n.pxlen--; ip##ipv##_clrbit(&nuc.ip##ipv.prefix, nuc.ip##ipv.pxlen); \
+  } \
+} while(0); return NULL;
 
-  case NET_IP6:
-    if (tab->trie)
-      return net_route_ip6_trie(tab, (net_addr_ip6 *) n);
-    else
-      return net_route_ip6_fib (tab, (net_addr_ip6 *) n);
+#define FVR_IP(ipv, var) \
+    net *r; if (r = net_find_valid(tab, (net_addr *) &var)) return r;
 
-  case NET_VPN6:
-    if (tab->trie)
-      return net_route_vpn6_trie(tab, (net_addr_vpn6 *) n);
-    else
-      return net_route_vpn6_fib (tab, (net_addr_vpn6 *) n);
+#define FVR_VPN(ipv, var) \
+    net_addr_vpn##ipv _var0 = NET_ADDR_VPN##ipv(var.prefix, var.pxlen, nu->vpn##ipv.rd); FVR_IP(ipv, _var0);
 
-  case NET_IP6_SADR:
-    if (tab->trie)
-      return net_route_ip6_sadr_trie(tab, (net_addr_ip6_sadr *) n);
-    else
-      return net_route_ip6_sadr_fib (tab, (net_addr_ip6_sadr *) n);
+  if (tab->trie)
+    switch (n->type) {
+      case NET_IP4:   TW(4, FVR_IP);
+      case NET_VPN4:  TW(4, FVR_VPN);
+      case NET_IP6:   TW(6, FVR_IP);
+      case NET_VPN6:  TW(6, FVR_VPN);
 
-  default:
-    return NULL;
-  }
-}
-
-
-static int
-net_roa_check_ip4_trie(rtable *tab, const net_addr_ip4 *px, u32 asn)
-{
-  int anything = 0;
-
-  TRIE_WALK_TO_ROOT_IP4(tab->trie, px, px0)
-  {
-    net_addr_roa4 roa0 = NET_ADDR_ROA4(px0.prefix, px0.pxlen, 0, 0);
-
-    struct fib_node *fn;
-    for (fn = fib_get_chain(&tab->fib, (net_addr *) &roa0); fn; fn = fn->next)
-    {
-      net_addr_roa4 *roa = (void *) fn->addr;
-      net *r = fib_node_to_user(&tab->fib, fn);
-
-      if (net_equal_prefix_roa4(roa, &roa0) && rte_is_valid(r->routes))
-      {
-	anything = 1;
-	if (asn && (roa->asn == asn) && (roa->max_pxlen >= px->pxlen))
-	  return ROA_VALID;
-      }
+      case NET_IP6_SADR:
+	return net_route_ip6_sadr_trie(tab, (net_addr_ip6_sadr *) n);
+      default:
+	return NULL;
     }
-  }
-  TRIE_WALK_TO_ROOT_END;
+  else
+    switch (n->type) {
+      case NET_IP4:   FW(4, FVR_IP);
+      case NET_VPN4:  FW(4, FVR_VPN);
+      case NET_IP6:   FW(6, FVR_IP);
+      case NET_VPN6:  FW(6, FVR_VPN);
 
-  return anything ? ROA_INVALID : ROA_UNKNOWN;
-}
-
-static int
-net_roa_check_ip4_fib(rtable *tab, const net_addr_ip4 *px, u32 asn)
-{
-  struct net_addr_roa4 n = NET_ADDR_ROA4(px->prefix, px->pxlen, 0, 0);
-  struct fib_node *fn;
-  int anything = 0;
-
-  while (1)
-  {
-    for (fn = fib_get_chain(&tab->fib, (net_addr *) &n); fn; fn = fn->next)
-    {
-      net_addr_roa4 *roa = (void *) fn->addr;
-      net *r = fib_node_to_user(&tab->fib, fn);
-
-      if (net_equal_prefix_roa4(roa, &n) && rte_is_valid(r->routes))
-      {
-	anything = 1;
-	if (asn && (roa->asn == asn) && (roa->max_pxlen >= px->pxlen))
-	  return ROA_VALID;
-      }
+      case NET_IP6_SADR:
+	return net_route_ip6_sadr_fib (tab, (net_addr_ip6_sadr *) n);
+      default:
+	return NULL;
     }
 
-    if (n.pxlen == 0)
-      break;
-
-    n.pxlen--;
-    ip4_clrbit(&n.prefix, n.pxlen);
-  }
-
-  return anything ? ROA_INVALID : ROA_UNKNOWN;
+#undef TW
+#undef FW
+#undef FVR_IP
+#undef FVR_VPN
 }
 
-static int
-net_roa_check_ip6_trie(rtable *tab, const net_addr_ip6 *px, u32 asn)
-{
-  int anything = 0;
-
-  TRIE_WALK_TO_ROOT_IP6(tab->trie, px, px0)
-  {
-    net_addr_roa6 roa0 = NET_ADDR_ROA6(px0.prefix, px0.pxlen, 0, 0);
-
-    struct fib_node *fn;
-    for (fn = fib_get_chain(&tab->fib, (net_addr *) &roa0); fn; fn = fn->next)
-    {
-      net_addr_roa6 *roa = (void *) fn->addr;
-      net *r = fib_node_to_user(&tab->fib, fn);
-
-      if (net_equal_prefix_roa6(roa, &roa0) && rte_is_valid(r->routes))
-      {
-	anything = 1;
-	if (asn && (roa->asn == asn) && (roa->max_pxlen >= px->pxlen))
-	  return ROA_VALID;
-      }
-    }
-  }
-  TRIE_WALK_TO_ROOT_END;
-
-  return anything ? ROA_INVALID : ROA_UNKNOWN;
-}
-
-static int
-net_roa_check_ip6_fib(rtable *tab, const net_addr_ip6 *px, u32 asn)
-{
-  struct net_addr_roa6 n = NET_ADDR_ROA6(px->prefix, px->pxlen, 0, 0);
-  struct fib_node *fn;
-  int anything = 0;
-
-  while (1)
-  {
-    for (fn = fib_get_chain(&tab->fib, (net_addr *) &n); fn; fn = fn->next)
-    {
-      net_addr_roa6 *roa = (void *) fn->addr;
-      net *r = fib_node_to_user(&tab->fib, fn);
-
-      if (net_equal_prefix_roa6(roa, &n) && rte_is_valid(r->routes))
-      {
-	anything = 1;
-	if (asn && (roa->asn == asn) && (roa->max_pxlen >= px->pxlen))
-	  return ROA_VALID;
-      }
-    }
-
-    if (n.pxlen == 0)
-      break;
-
-    n.pxlen--;
-    ip6_clrbit(&n.prefix, n.pxlen);
-  }
-
-  return anything ? ROA_INVALID : ROA_UNKNOWN;
-}
 
 /**
  * roa_check - check validity of route origination in a ROA table
@@ -523,22 +294,56 @@ net_roa_check_ip6_fib(rtable *tab, const net_addr_ip6 *px, u32 asn)
 int
 net_roa_check(rtable *tab, const net_addr *n, u32 asn)
 {
+  net_addr_union *nu = SKIP_BACK(net_addr_union, n, n);
+  int anything = 0;
+  struct fib_node *fn;
+
+#define TW(ipv) do {								\
+  TRIE_WALK_TO_ROOT_IP##ipv(tab->trie, &(nu->ip##ipv), var) {			\
+    net_addr_roa##ipv roa0 = NET_ADDR_ROA##ipv(var.prefix, var.pxlen, 0, 0);	\
+    ROA_PARTIAL_CHECK(ipv);							\
+  } TRIE_WALK_TO_ROOT_END;							\
+  return anything ? ROA_INVALID : ROA_UNKNOWN;					\
+} while (0)
+
+#define FW(ipv) do {								\
+  net_addr_roa##ipv roa0 = NET_ADDR_ROA##ipv(nu->ip##ipv.prefix, nu->ip##ipv.pxlen, 0, 0);\
+  while (1) {									\
+    ROA_PARTIAL_CHECK(ipv);							\
+    if (roa0.pxlen == 0) break;							\
+    roa0.pxlen--; ip##ipv##_clrbit(&roa0.prefix, roa0.pxlen);			\
+  }										\
+} while (0)
+
+#define ROA_PARTIAL_CHECK(ipv) do {						\
+  for (fn = fib_get_chain(&tab->fib, (net_addr *) &roa0); fn; fn = fn->next)	\
+  {										\
+    net_addr_roa##ipv *roa = (void *) fn->addr;				\
+    net *r = fib_node_to_user(&tab->fib, fn);					\
+    if (net_equal_prefix_roa##ipv(roa, &roa0) && rte_is_valid(r->routes))	\
+    {										\
+      anything = 1;								\
+      if (asn && (roa->asn == asn) && (roa->max_pxlen >= nu->ip##ipv.pxlen))	\
+      return ROA_VALID;							\
+    }										\
+  }										\
+} while (0)
+
   if ((tab->addr_type == NET_ROA4) && (n->type == NET_IP4))
   {
-    if (tab->trie)
-      return net_roa_check_ip4_trie(tab, (const net_addr_ip4 *) n, asn);
-    else
-      return net_roa_check_ip4_fib (tab, (const net_addr_ip4 *) n, asn);
+    if (tab->trie)  TW(4);
+    else	    FW(4);
   }
   else if ((tab->addr_type == NET_ROA6) && (n->type == NET_IP6))
   {
-    if (tab->trie)
-      return net_roa_check_ip6_trie(tab, (const net_addr_ip6 *) n, asn);
-    else
-      return net_roa_check_ip6_fib (tab, (const net_addr_ip6 *) n, asn);
+    if (tab->trie)  TW(6);
+    else	    FW(6);
   }
-  else
-    return ROA_UNKNOWN;	/* Should not happen */
+
+  return anything ? ROA_INVALID : ROA_UNKNOWN;
+#undef ROA_PARTIAL_CHECK
+#undef TW
+#undef FW
 }
 
 /**
@@ -1748,6 +1553,9 @@ rt_examine(rtable *t, net_addr *a, struct channel *c, const struct filter *filte
 void
 rt_refresh_begin(rtable *t, struct channel *c)
 {
+  if (c->debug & D_EVENTS)
+    log(L_TRACE "%s.%s: Route refresh begin", c->proto->name, c->name);
+
   FIB_WALK(&t->fib, net, n)
     {
       rte *e;
@@ -1769,6 +1577,9 @@ rt_refresh_begin(rtable *t, struct channel *c)
 void
 rt_refresh_end(rtable *t, struct channel *c)
 {
+  if (c->debug & D_EVENTS)
+    log(L_TRACE "%s.%s: Route refresh end", c->proto->name, c->name);
+
   int prune = 0;
 
   FIB_WALK(&t->fib, net, n)
@@ -2148,6 +1959,7 @@ rt_setup(pool *pp, struct rtable_config *cf)
   t->name = cf->name;
   t->config = cf;
   t->addr_type = cf->addr_type;
+  t->debug = cf->debug;
 
   fib_init(&t->fib, p, t->addr_type, sizeof(net), OFFSETOF(net, n), 0, NULL);
 
@@ -2860,6 +2672,7 @@ rt_new_table(struct symbol *s, uint addr_type)
   c->gc_period = (uint) -1;	/* set in rt_postconfig() */
   c->min_settle_time = 1 S;
   c->max_settle_time = 20 S;
+  c->debug = new_config->table_default_debug;
 
   add_tail(&new_config->tables, &c->n);
 
@@ -2917,6 +2730,7 @@ rt_reconfigure(rtable *tab, struct rtable_config *new, struct rtable_config *old
   new->table = tab;
   tab->name = new->name;
   tab->config = new;
+  tab->debug = new->debug;
 
   return 1;
 }
@@ -3663,6 +3477,7 @@ rt_get_hostentry(rtable *tab, ip_addr a, ip_addr ll, rtable *dep)
       return he;
 
   he = hc_new_hostentry(hc, tab->rp, a, link, dep, k);
+  he->owner = tab;
   rt_update_hostentry(tab, he);
   return he;
 }
