@@ -172,6 +172,17 @@ struct bfd_session
   u32 tx_csn_time;			/* Timestamp of last tx_csn change */
 };
 
+struct bfd_show_sessions_cmd {
+  net_addr address;
+  struct iface *iface;
+  struct symbol *name;
+  u8 verbose;
+  u8 ipv4;
+  u8 ipv6;
+  u8 direct;
+  u8 multihop;
+};
+
 
 extern const char *bfd_state_names[];
 
@@ -218,7 +229,7 @@ static inline void bfd_unlock_sessions(struct bfd_proto *p) { pthread_spin_unloc
 struct bfd_session * bfd_find_session_by_id(struct bfd_proto *p, u32 id);
 struct bfd_session * bfd_find_session_by_addr(struct bfd_proto *p, ip_addr addr, uint ifindex);
 void bfd_session_process_ctl(struct bfd_session *s, u8 flags, u32 old_tx_int, u32 old_rx_int);
-void bfd_show_sessions(struct proto *P);
+void bfd_show_sessions(struct proto *P, struct bfd_show_sessions_cmd *args);
 
 /* packets.c */
 void bfd_send_ctl(struct bfd_proto *p, struct bfd_session *s, int final);
